#include <cstddef>
#include "AST.h"
#include "CfSource.h"
#include "CodeContext.h"
#include "parse.h"


const int CodeContext::recoverSlot = ios::xalloc();


inline CodeContext *CodeContext::recover( const ostream &out )
{
  void * const pointer = out.pword( recoverSlot );
  return static_cast< CodeContext * >( pointer );    
}

static string currLineDirective(const CodeContext *context)
{
  extern bool line_directives;
  if (line_directives && context->isValidDebug())
    return string("\n#line ") +
      int2str(context->getCurrDebugLine()) +
      " \"" +
      *context->getCurrDebugFile()->name() +
      "\"" +
      '\n';
  else
    return "\n";
}

static string startLineDirective(const CodeContext *context)
{
  extern bool line_directives;
  if (line_directives && context->isValidDebug())
    return string("\n#line ") +
      int2str(context->getStartDebugLine()) +
      " \"" +
      *context->getStartDebugFile()->name() +
      "\"" +
      '\n';
  else
    return "\n";
}

ostream &endCline( ostream &outs )
{
  CodeContext &context = *CodeContext::recover( outs );
  
  outs << currLineDirective(&context);
  return outs;
}


CodeContext::CodeContext( CodeContext &parent, ostream &sink )
  : LocalVars( parent ),
    sink( sink ),
    parent( &parent ),
    strings( parent.strings ),
    lastNodeSeen( parent.lastNodeSeen ),
    curr_debug_line( parent.curr_debug_line ),
    curr_debug_file( parent.curr_debug_file ),
    start_debug_line( parent.curr_debug_line ),
    start_debug_file( parent.curr_debug_file ),
    level( parent.level + 1 )
{
  pword( recoverSlot ) = this;
}


CodeContext::CodeContext( CodeContext &parent )
  : LocalVars( parent ),
    sink( parent ),
    parent( &parent ),
    strings( parent.strings ),
    lastNodeSeen( parent.lastNodeSeen ),
    curr_debug_line( parent.curr_debug_line ),
    curr_debug_file( parent.curr_debug_file ),
    start_debug_line( parent.curr_debug_line ),
    start_debug_file( parent.curr_debug_file ),
    level( parent.level + 1 )
{
  pword( recoverSlot ) = this;
}


CodeContext::CodeContext( CfSource &file )
  : LocalVars( file ),
    sink( file ),
    strings( file ),
    parent( 0 ),
    lastNodeSeen( 0 ),
    curr_debug_line( -1 ),
    curr_debug_file( 0 ),
    start_debug_line( -1 ),
    start_debug_file( 0 ),
    level( 0 )
{
  pword( recoverSlot ) = this;
}


CodeContext::~CodeContext()
{
  codeGen();

  if (parent != NULL) {
    if (lastNodeSeen != NULL) {
      parent->setDebugLine(lastNodeSeen);
    }
  }
}


void CodeContext::codeGen() const
{
  codeGen( sink );
}

void CodeContext::setDebugLine(TreeNode *t)
{
  const SourcePosn position = t->position();

  if ((debugger_level > 0) && (!position.unknown())) {
    curr_debug_line = position.posnToLineNumber();
    curr_debug_file = position.file;
    (*this) << endCline;
    if (start_debug_line == -1) {
      start_debug_line = curr_debug_line;
      start_debug_file = curr_debug_file;
    }
    lastNodeSeen = t;
  } else {
    /* Do nothing */
  }
}

void CodeContext::codeGen( ostream &sink ) const
{
  extern bool line_directives;
  sink << startLineDirective(this);
  sink << "{ /* Level: " << level << " */";
  sink << startLineDirective(this);

  LocalVars::codeGen( sink, start_debug_line, start_debug_file);

  sink << const_cast< CodeContext * >( this )->rdbuf()
       << "} /* Level: " << level << " */";
  sink << currLineDirective(this);
}
