#include "utils.h"
#include "AST.h"
#include "ClassDecl.h"
#include "CodeContext.h"
#include "CtExternal.h"
#include "CtGlobal.h"
#include "CtLocal.h"
#include "PrimitiveDecl.h"
#include "code.h"
#include "decls.h"
#include "delimit.h"
#include "jni-bridge.h"
#include "lgMacro.h"


void ThisConstructorCallNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    vector< string > parameters;
    parameters.push_back( "me" );
    args()->emitExpressionList( context, parameters );

    context << decl()->cMethodNameStatic() << '('
	    << delimit( parameters )
	    << ");" << endCline;
}

void SuperConstructorCallNode::codeGen( CodeContext &context )
{
    context.setDebugLine(this);

    ClassDecl * const caller = static_cast< ClassDecl * >(parent()->parent()->parent()->decl());
    assert( caller->category() == Decl::Class );
    
    if (caller != ObjectDecl) {
	
	vector< string > parameters;
	ClassDecl &super = static_cast< ClassDecl & >(*decl()->container());
	assert( super.category() == Decl::Class );
	
	parameters.push_back('(' + super.cType().name + " *) me");
	args()->emitExpressionList( context, parameters );
	
	context << decl()->cMethodNameStatic() << '('
		<< delimit( parameters )
		<< ");" << endCline;
    }
    
    // The rules for chaining one constructor onto another
    // constructor for the same class ultimately mean that
    // only a subset of "terminal" constructors will call
    // the superclass constructor.  The same subset can also
    // bear responsibility for intializing instance fields
    // that were declared with initializer expressions.
    context << caller->cInitFieldsName() << "( me );" << endCline;
}

void StaticInitNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    // semantics??
    block()->codeGen( context );
}


void EmptyStmtNode::codeGen( CodeContext & context)
{
  context.setDebugLine(this);
}

void BlockNode::emitStatement( CodeContext &context )
{
    CodeContext subcontext( context );
    stmts()->codeGen( subcontext );

    subcontext.setDebugLine(this);
}

void LabeledStmtNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);

    stmt()->codeGen( context );
}

void IfStmtNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);

    const string predicate = condition()->emitExpression( context );
    context << "if (" << predicate << ") ";
    {
	CodeContext thenContext( context );
	thenPart()->codeGen( thenContext );
    }
    if (!elsePart()->absent()) {
	context << "else ";
	CodeContext elseContext( context );
	elsePart()->codeGen( elseContext );
    }
}

void WhileNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);

    const string predicate = test()->declareTemporary( context );
    context << "do ";
    {
	CodeContext subcontext( context );

	emitContinueTarget( subcontext );
	const string predicateEval = test()->emitExpression( subcontext );
	subcontext << predicate << " = " << predicateEval << ";" << endCline;
	subcontext << "if (" << predicate << ") ";
	{
	    CodeContext bodyContext( subcontext );
	    stmt()->codeGen( bodyContext );
	}
    }
    context << "while (" << predicate << ");" << endCline;
#if 0
    context << "{ ";
    {
	CodeContext subcontext( context );
	const string predicate = test()->declareTemporary( subcontext );

	emitContinueTarget( subcontext );
	const string predicateEval = test()->emitExpression( subcontext );
	subcontext << predicate << " = " << predicateEval << ";" << endCline;
	subcontext << "while (" << predicate << ") ";
	{
	    CodeContext bodyContext( subcontext );
	    stmt()->codeGen( bodyContext );
	}
    }
    context << "} " << endCline;
#endif
}

void DoNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);

    const string predicate = test()->declareTemporary( context );
    context << "do ";
    {
	CodeContext subcontext( context );
	stmt()->codeGen( subcontext );
	emitContinueTarget( subcontext );
	const string predicateEval = test()->emitExpression( subcontext );
	subcontext << predicate << " = " << predicateEval << ";" << endCline;
    }
    context << "while (" << predicate << ");" << endCline;
}

void ForNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);

    bool testAbsent = test()->absent();

    // Must contain local variable declarations
    // that the init portion may introduce.
    CodeContext initContext( context );
    init()->codeGen( initContext );
    
    initContext << "do ";
    {
	CodeContext subcontext( initContext );
	
	if (!testAbsent) {
	    const string predicateEval = test()->emitExpression( subcontext );
	    subcontext << "if (" << predicateEval << ") {" << endCline;
	}
	
	stmt()->codeGen( subcontext );
	emitContinueTarget( subcontext );
	update()->codeGen( subcontext );
	
	if (!testAbsent)
	    subcontext << "} else break;" << endCline;
    }
    initContext << "while (1);" << endCline;
}

void SynchronizedNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);
  CodeContext subcontext(context);

  static const CtExternal monitor( "titanium_monitor_t" );
  static const CtLocal localMonitor( monitor );
  static const CtGlobal globalMonitor( monitor );
  static const CtGlobal globalLocalMonitor( localMonitor );

  string proc;
  const string mon = ExprNode::declareTemporary(subcontext, localMonitor);

  subcontext << "#ifdef HAVE_MONITORS" << endCline;

  if (expr()->absent()) {
    proc = "0";
    const string monAddr = ExprNode::declareTemporary(subcontext, globalLocalMonitor);
    subcontext << "TO_GLOBAL(" << monAddr << ", 0, &class_monitor);" << endCline
	       << "DEREF_GLOBAL(" << mon << ", " << monAddr << ");" << endCline
	       << "#else" << endCline
	       << mon << " = class_monitor;" << endCline;
  } else {
    const string mutex = expr()->emitExpression(subcontext);
    if (expr()->type()->isLocal()) {
      proc = "MYPROC";
      subcontext << mon << " = " << mutex << "->monitor;\n" << endCline;
    } else {
      const string monAddr = ExprNode::declareTemporary(subcontext, globalLocalMonitor);
      proc = ExprNode::declareTemporary(subcontext, PrimitiveDecl::IntDecl.cType());
      
      subcontext << proc << " = TO_PROC(" + mutex + ");" << endCline
		 << "FIELD_ADDR_GLOBAL(" << monAddr << ", " << mutex << ", monitor);" << endCline
		 << "DEREF_GLOBAL(" << mon << ", " << monAddr << ");" << endCline;
    }
  }
  subcontext << "#endif" << endCline;
  
  string monitorexit = "monitorExit( " + mon + ", " + proc + " );";
  
  subcontext << "monitorEnter( " << mon << ", " << proc << " );" << endCline;
  TryStmtNode* trystatement = new TryStmtNode(new TryNode(stmt()),
					      NULL,
					      new FinallyNode(new CodeLiteralNode(monitorexit)));   
  child(1, trystatement);
  trystatement->codeGen(subcontext);
}

void CodeLiteralNode::codeGen ( CodeContext &context ) {
  context.setDebugLine(this);
  context << codeString() << endCline;
}

void ThrowNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);
  string exception = expr()->emitExpression( context );  
  context << "THROW(" << exception << ");" << endCline;
}


void ParameterNode::codeGen( CodeContext & context)
{
  context.setDebugLine(this);

    // dont worry (handled by methodDeclNode)
}


void NameNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    if (!qualifier()->absent()) {
	qualifier()->codeGen( context );
	context << "_" << *ident();
    } else
	context << *ident();
}


void PartitionStmtNode::emitStatement( CodeContext & context)
{
  context.setDebugLine(this);
}
