#include <map>
#include <set>
#include "AST.h"
#include "StringSet.h"
#include "code-util.h"

extern bool debug_defs;
#define debug_pure debug_defs

// Returns whether s appears in z, a list of n strings.
bool strFind(const char *z[], int n, const string& s)
{
  static map< const char **, StringSet *, less< const char ** > > memo;
  StringSet *& S = memo[z];

  if (S == NULL) {
    S = new StringSet;
    while (--n >= 0) S->insert(z[n]);
  }
  return (S->count(s) > 0);
} 

/* If t is a call to binary +, -, *, /, or %, then return the op (e.g., '+').
   Otherwise return '\0'. */
char arithCall(const TreeNode *t) 
{
  char *s = "+-*/%";

  if (isOFAN(t->child(0)) && isNameNode(t->child(0)->child(1))) {
    const TreeNode *namenode = t->child(0)->child(1);
    if (!namenode->qualifier()->absent()) return false;
    const string methodName = *namenode->ident();
    for (int i = 0; ; i++)
      if (s[i] == '\0'
	  || methodName.c_str()[0] == s[i] && methodName.c_str()[1] == '\0')
	return s[i];
  }
  return false;
}

bool tiOrNumericType(const TreeNode *x)
{
  ExprNode *t = (ExprNode *) x;
  return (t->type()->isArithType()
	  || t->type()->isPointType()
	  || t->type()->isDomainType()
	  || t->type()->isRectDomainType()
	  || t->type()->isTitaniumArrayType());
}

/////////////////////////////////////////////////////////////////////////////

/*
   Terminology:

   A "side-effect free" method has no side effects that are visible
   to the callee.

   A pure method always returns the same value for the same arguments
   and is "side-effect free."

*/

static const char *PointStaticPureMethods[] = { "all", "direction" };
static const char *RectDomainStaticPureMethods[] = {};
static const char *DomainStaticPureMethods[] = { "toDomain" };
static const char *GridStaticPureMethods[] = {};

static const char *PointPureMethods[] = { "arity", "+", "-", "*", "/",
"<", ">", "<=", ">=", "==", "!=" };

static const char *GridPureMethods[] = { "domain", "arity",
"translate", "restrict", "inject", "project", "slice", "transpose" };

static const char *RectDomainPureMethods[] = { "isRectangular",
"arity", "min", "max", "boundingBox", "stride", "size", "isNull",
"contains", "accrete", "shrink", "border", "slice", "+", "-", "*", "/",
"<", ">", "<=", ">=", "==", "!=" };

static const char *DomainPureMethods[] = { "isRectangular", "arity",
"min", "max", "boundingBox", "stride", "size", "isNull", "contains",
"RectDomainList", "PointList", "+", "-", "*", "/", "<", ">", "<=",
">=", "==", "!=" };

static const char *PureMethods[] = {
  "Math.abs", "Math.round", "Math.max", "Math.min", "Math.sin", "Math.cos",
  "Math.tan", "Math.asin", "Math.acos", "Math.atan", "Math.exp", "Math.log",
  "Math.sqrt", "Math.IEEEremainder", "Math.ceil", "Math.floor", "Math.rint",
  "Math.atan2", "Math.pow",

  "Ti.thisProc", "Ti.numProcs"
};

static const char *SideEffectFreeMethods[] = {"Math.random"};

static bool methodNameFind2(TreeNode *t, const char *strs[], int n)
{
  assert(isNameNode(t));
  return strFind(strs, n, *t->ident());
}

#define methodNameFind(node, strs) \
	(methodNameFind2((node), strs, sizeof(strs) / sizeof(strs[0])))

#define strFind1(str, strs) \
	strFind((strs), sizeof(strs) / sizeof((strs)[0]), (str))

bool TreeNode::isSideEffectFree()
{
  undefined("isSideEffectFree");
  return false;
}

// Return true if method is a pure function or if its name appears in
// SideEffectFreeMethods.
bool MethodCallNode::isSideEffectFree()
{
  TreeNode *m = method();
  bool r = isPureFunction();
  string s;
  TypeNode *type;
  if (!r && (isTypeFAN(m) && ((type = m->ftype()) != NULL)
	     || isOFAN(m) && isThisNode(m->object())
	     && ((type = m->object()->type()) != NULL))) {
    s = *type->name()->simpName()->ident()
      + '.' + *m->simpName()->ident();
    r = strFind1(s, SideEffectFreeMethods);
  }
  if (debug_pure)
    cout << "sideEffectFree(" << s << ") => " << (r ? "yes" : "no") << '\n';

  return r;
}

bool TreeNode::isPureFunction()
{
  undefined("isPureFunction");
  return false;
}

bool MethodCallNode::isPureFunction()
{
  // cout << "                    isPureFunction():";
  // print(cout, 20);
  // cout << "\n";

  TreeNode *m = method();
  if (isTypeFAN(m)) {
    TypeNode *accessedObjectType = m->accessedObjectType();
    if (accessedObjectType->isPointType())
      return methodNameFind(m->simpName(), PointStaticPureMethods);
    else if (accessedObjectType->isRectDomainType())
      return methodNameFind(m->simpName(), RectDomainStaticPureMethods);
    else if (accessedObjectType->isDomainType())
      return methodNameFind(m->simpName(), DomainStaticPureMethods);
    else if (accessedObjectType->isTitaniumArrayType())
      return methodNameFind(m->simpName(), GridStaticPureMethods);
    else if (isTypeNameNode(m->ftype())) {
      string s = *m->ftype()->name()->simpName()->ident()
	+ '.' + *m->simpName()->ident();
      bool r = strFind1(s, PureMethods);
      if (debug_pure)
	cout << "pure(" << s << ") => " << (r ? "yes" : "no") << '\n';
      return r;
    }
  } else if (isOFAN(m)) {
    TypeNode *accessedObjectType = m->accessedObjectType();
    if (accessedObjectType->isPointType())
      return methodNameFind(m->simpName(), PointPureMethods);
    else if (accessedObjectType->isRectDomainType())
      return methodNameFind(m->simpName(), RectDomainPureMethods);
    else if (accessedObjectType->isDomainType())
      return methodNameFind(m->simpName(), DomainPureMethods);
    else if (accessedObjectType->isTitaniumArrayType())
      return methodNameFind(m->simpName(), GridPureMethods);
  } else if (arithCall(this)) {
    TreeNode *left = child(0)->child(0);
    TreeNode *right = child(1)->child(0);
    if (tiOrNumericType(left) && tiOrNumericType(right)) return true;
  }

  return false;
}

bool isGridMethod(MethodCallNode *t)
{
  TreeNode *m = t->method();
  return isOFAN(m) && m->object()->type()->isTitaniumArrayType();
}

bool isGridMethod_domain(TreeNode *t)
{
  MethodCallNode *m;
  ObjectFieldAccessNode *o;

  return isMethodCallNode(t) && isGridMethod(m = ((MethodCallNode *) t))
    && isOFAN(m->method())
    && ((o = ((ObjectFieldAccessNode *) m->method())))
    && (*(o->simpName()->ident()) == "domain");
}

/* Assumes isGridMethod(t) is true. */
bool isGridMethodMutating(MethodCallNode *t)
{
  return !t->isPureFunction();
}



/////////////////////////////////////////////////////////////////////////////

bool isMethodCallNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "MethodCallNode") == 0);
}  

bool isPointNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "PointNode") == 0);
}  

bool isOFAN(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "ObjectFieldAccessNode") == 0);
}  

bool isTypeFAN(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "TypeFieldAccessNode") == 0);
}  

bool isThisFAN(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "ThisFieldAccessNode") == 0);
}  

bool isSFAN(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "SuperFieldAccessNode") == 0);
}  

bool isNameNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "NameNode") == 0);
}  

bool isTypeNameNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "TypeNameNode") == 0);
}  

bool isThisNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "ThisNode") == 0);
}  

bool TreeNode::isArrayAccessNode() const
{
  return false;
}  

bool ArrayAccessNode::isArrayAccessNode() const
{
  return true;
}  

bool isObjectNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "ObjectNode") == 0);
}  

bool isForEachStmtNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "ForEachStmtNode") == 0);
}  

bool isForEachPairNode(const TreeNode *t)
{
  return  (strcmp(t->oper_name(), "ForEachPairNode") == 0);
}  

bool isMethodDeclNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "MethodDeclNode") == 0);
}

bool isTreeListNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "*list*") == 0);
}

bool isClassDeclNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "ClassDeclNode") == 0);
}

bool isCompileUnitNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "CompileUnitNode") == 0);
}

bool isBlockNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "BlockNode") == 0);
}

bool isOmittedNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "-*-") == 0);
}

bool isVarDeclNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "VarDeclNode") == 0);
}

bool isCastNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "CastNode") == 0);
}

bool isAssignNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "AssignNode") == 0);
}

bool isSynchronizedNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "SynchronizedNode") == 0);
}

bool isParameterNode(const TreeNode *t) 
{
  return (strcmp(t->oper_name(), "ParameterNode") == 0);
}

bool StatementNode::isStatementNode(void) {
  return true;
}

