// -*-C++-*-
// compiler.h: general compiler definitions

#ifndef _COMPILER_H
#define _COMPILER_H

#include <string>
#include <stdio.h>
#include "llist.h"
#include "stats.h"

class ClassDecl;
class Decl;
class TreeNode;

extern int DEBUG;
// Set to 1 to get some debugging output

extern int debugger_level;
// Set to 1 to output information for a debugger

extern string dump_ast;
// Set to a filename to dump AST there

extern llist<TreeNode *> *allFiles;
// The list of all CompileUnitNodes loaded during compilation

void initTheSystem();
// Effects: initialise the compiler. Call before doing anything else

void load(const string name, bool titaniumFile);
// Effects: Opens and parses a file called 'name', and performs
// 'nameResolution' (build the file-level environment, load() all
// types explicitly named)
// If titaniumFile is true, the file is assumed to contain titanium source code
// if it is false, only Java code is allowed.

void load(const char * const name, bool titaniumFile);
// Effects: Opens and parses a file called 'name', and performs
// 'nameResolution' (build the file-level environment, load() all
// types explicitly named)
// If titaniumFile is true, the file is assumed to contain titanium source code
// if it is false, only Java code is allowed.

void load(const string name, FILE *file, bool titaniumFile);
// Effects: Parses file, assuming it is a file called 'name', and
//   performs 'nameResolution' (build the file-level environment,
//   load() all types explicitly named)
// If titaniumFile is true, the file is assumed to contain titanium source code
// if it is false, only Java code is allowed.

void staticSemantics();
// Effects: Run the static semantics on the currently loaded files.
//   Note that this may cause more files to be loaded...

bool isTitanium(string fileName);
// Returns: true if fileName ends in ".ti"

void foreachOptPass();
// Searches tree for index expressions that may be optimizable.
// See code-foreach.cc.

void codeGen();
// Backend


// Some phases of the front end

void buildEnvironments();
// Effects: builds the class-level environments for all the classes loaded
// since the last call to buildEnvironments

void verifyCircularity();
// Effects: Report errors for all classes or interfaces that are in a circular
//   inheritance dependency

void foldConstants();
// Effects: Perform global constant folding

void singleAnalysis(TreeNode *file);
// Effects: Perform single analysis on the CompileUnitNode file

				/* Standard Declarations */

extern Decl *unnamedPackage;	/* The unnamed package */

extern ClassDecl
   *ObjectDecl,			/* java.lang.Object */
   *RuntimeExceptionDecl,	/* java.lang.RuntimeException */
   *ThrowableDecl,		/* java.lang.Throwable */
   *JavaArrayDecl,		/* ti.internal.tiJArray */
   *TiArrayDecl,		/* ti.internal.tiArray */
   *TiArrayLDecl,		/* ti.internal.tiArray */
   *TiArrayM1Decl,		/* ti.internal.tiArrayM1 */
   *TemplateArgDecl,		/* ti.internal.tiTemplateArgument */
   *PointDecl,			/* ti.internal.tiPoint */
   *DomainDecl,			/* ti.internal.tiDomain */
   *RectDomainDecl,		/* ti.internal.tiRectDomain */
   *RectDomainM1Decl,		/* ti.internal.tiRectDomainM1 */
   *ErrorDecl,			/* java.lang.Error */
   *CloneableDecl,		/* java.lang.Cloneable */
   *StringDecl,			/* Java.lang.String */
   *StringBufferDecl,		/* java.lang.StringBuffer */
   *RegionDecl,			/* ti.lang.Region */
   *SharedRegionDecl,		/* ti.lang.SharedRegion */
   *PrivateRegionDecl,		/* ti.lang.PrivateRegion */
   *NativeUtilsDecl;		/* ti.lang.NativeUtils */

extern Decl *TiArityDecl;	/* ti.internal.tiTemplateArgument.tiArity */


// Miscellaneous strings
extern const string
  *ObjectString,
  *UnknownString,
  *StringString,
  *StringBufferString,
  *ErrorString,
  *RuntimeExceptionString,
  *ThrowableString,
  *JavaArrayString,
  *RectDomainString,
  *RectDomainM1String,
  *DomainString,
  *TiArrayString,
  *TiArrayM1String,
  *TemplateArgString,
  *PointString,
  *CloneableString,
  *TitaniumString,
  *InternalString,
  *JavaString,
  *LangString,
  *ArityString,
  *RegionString,
  *SharedRegionString,
  *PrivateRegionString;

#endif
