#include <split-c.h>
#include <am.h>

/* 
   This file contains AM-2 based code that performs scatter/gather operations.
   These are designed to work with 2-and-3-dimensional arrays, but they can
   be called multiple times to handle more complex structures.
 */

static volatile ___globPtrType remotePacked;
static unsigned char *scatterBuf;
static handler_t bulkGather2HandlerIdx, bulkAllocReplyHandlerIdx, bulkFreeHandlerIdx,
		 bulkNullHandlerIdx, bulkScatter2HandlerIdx, bulkAllocHandlerIdx;

/* 
   Gathers N strips of remote data starting at global address addr, each of 
   length stripSize, with a "skip" after each strip. The skip is the number 
   of elements (to be skipped), which lie between the end of a strip and the 
   beginning of the next one. Used to gather a 2-D rectangular chunk of an 
   array. The data is placed in the storage pointed to by "dest". This is a 
   blocking call.
*/
void bulkGather2(___globPtrType addr, int stripSize, int N, int skip,
		 unsigned char *dest) {
  /* The strategy here is to make the remote node allocate a buffer,
     gather the data into that buffer,
     have it pass back a Split-C pointer to that data, 
     bulk-copy that data over here, and then have the remote node free 
     the data. This stuff could be streamlined a bit by having the remote node
     reply with the actual data instead of a pointer to it, but then we run
     into the AM-2 message size limit. So it's either I write fragmentation 
     code, or I leverage off the Split-C bulk transfer stuff at the cost of
     some (unknown quantity of) performance. */
  int remProc;
  unsigned char *remAddr;
  int i;
  remProc = GET_PROC(addr);
  remAddr = (unsigned char *)GET_ADDR(addr);

  /* Local gather. */
  if (remProc == ___MYPROC) {
    unsigned char *src;
    int strip;
    src = remAddr;
    for (strip = 1; strip <= N; strip++) {
      memcpy(dest, src, stripSize);
      dest += stripSize; src += stripSize + skip;
    }
    return;
  }

  /* Send a gather request to the remote node. */
  ((___splitPtrAsUnion*)(&remotePacked))->asStruct.addr = (void *)-1;
  AM_Safe(AM_Request4(__sc_endpoint, remProc, bulkGather2HandlerIdx, 
		      (int)remAddr, stripSize, N, skip));
  /* Wait for the reply to put the SC ptr to the packed data in 
     remotePacked. */
  while (GET_ADDR(remotePacked) == (void *)-1) {
    AM_Safe(AM_Poll(__sc_bundle));
  }
  /* Bulk-copy this data to local memory. */
  bulk_read(dest, remotePacked, stripSize * N);
  /* Send a free request to the endpoint. */
  AM_Safe(AM_Request4(__sc_endpoint, remProc, bulkFreeHandlerIdx, 
		      (int)GET_ADDR(remotePacked), 0, 0, 0));
}

/* Handles a gather request. Allocates a buffer, gathers the data into it, 
   and tells the requestor where the data is. The requestor must tell the
   gathering process to free the buffer when it's done using it. */
static void bulkGather2Handler(void *token, int addr, int stripSize,
			       int N, int skip)
{
  unsigned char *start = (unsigned char *)addr;
  unsigned char *buf;
  unsigned char *dest, *src;
  int strip;
  buf = (unsigned char *)malloc(stripSize * N);
  dest = buf; src = start;
  for (strip = 1; strip <= N; strip++) {
    memcpy(dest, src, stripSize);
    dest += stripSize; src += stripSize + skip;
  }
  /* Send a reply to the requestor with the buffer's proc number and address. 
   */
  AM_Safe(AM_Reply4(token, bulkAllocReplyHandlerIdx, ___MYPROC, (int)buf, 0, 
		    0));
}

#define GO_AWAY (-666)

/* Performs a scatter from local memory to a location described by a Split-C
   pointer. The parameters are as described above bulkGather2(). */
/* There's some trickiness here described above bulkAllocHandler(). */
void bulkScatter2(___globPtrType addr, unsigned char *buf, int stripSize,
		  int N, int skip) {
  int remProc, strip;
  unsigned char *dest;
  int done;

  dest = GET_ADDR(addr);
  remProc = GET_PROC(addr);
  if (remProc == ___MYPROC) {
    unsigned char *src;
    src = buf;
    for (strip = 1; strip <= N; strip++) {
      memcpy(dest, src, stripSize);
      src += stripSize;
      dest += stripSize + skip;
    }
    return;
  }

  /* Allocate a scatter buffer on the remote process. */
  done = 0;
  while (!done) {
    /* Set remotePacked to a bogus value. */
    ((___splitPtrAsUnion*)(&remotePacked))->asStruct.addr = (void *)-1;
    AM_Safe(AM_Request4(__sc_endpoint, remProc, bulkAllocHandlerIdx,
			stripSize * N, 0, 0, 0));
    /* Wait for the remote to run the allocation handler. */
    while (GET_ADDR(remotePacked) == (void *)-1) {
      AM_Safe(AM_Poll(__sc_bundle));
    }
    /* The allocation could fail if someone else is already doing a scatter. */
    if (GET_ADDR(remotePacked) != (unsigned char *)GO_AWAY) {
      done = 1;
    }
  }    

  /* Push the data into the scatter buffer. */
  bulk_write(remotePacked, buf, stripSize * N);
  /* Tell the remote process to do the scatter and free up its scatter 
     buffer.  */
  AM_Safe(AM_Request4(__sc_endpoint, remProc, bulkScatter2HandlerIdx, 
		      (int)dest, stripSize, N, skip));
}

/* Someone asked us to allocate a scatter buffer with the given size.
   Reply back with its process number and address. We are responsible for
   freeing this storage after we do the scatter. */
static void bulkAllocHandler(void *token, int size, int dummy0, int dummy1,
			     int dummy2) {
  /* We need to guard against multiple processes telling us to allocate a
     scatter buffer. We'll just tell them to go away if the scatter buffer's
     being used. The other solution is to have a list of buffers, but that
     could chew up a lot of memory, and it's unclear what this would gain. */
  if (!scatterBuf) {
    scatterBuf = (unsigned char *)malloc(size);
    AM_Safe(AM_Reply4(token, bulkAllocReplyHandlerIdx, ___MYPROC, 
		      (int)scatterBuf, 0, 0));
  }
  else {
    AM_Safe(AM_Reply4(token, bulkAllocReplyHandlerIdx, ___MYPROC,
		      GO_AWAY, 0, 0));
  }
}

/* Handles a scatter request. */
static void bulkScatter2Handler(void *token, int addr, int stripSize,
				int N, int skip, int dummy0) {
  int strip;
  unsigned char *dest, *src;
  src = scatterBuf;
  dest = (unsigned char *)addr;
  for (strip = 1; strip <= N; strip++) {
    memcpy(dest, src, stripSize);
    src += stripSize;
    dest += stripSize + skip;
  }
  free(scatterBuf);
  scatterBuf = 0;
  AM_Safe(AM_Reply4(token, bulkNullHandlerIdx, 0, 0, 0, 0));
}


/* Some remote process has just allocated memory, and is giving us a pointer
   to it. We put it in Split-C pointer form, in remotePacked. */
static void bulkAllocReplyHandler(void *token, int proc, int addr,
				  int dummy0, int dummy1) {
  remotePacked = ___create_global_ptr(proc, (unsigned char *)addr);
}

/* Frees memory at the given location, which was earlier allocated with 
   malloc(). Would be nice if SC had a bulk-read-and-free thing, but oh
   well. */
static void bulkFreeHandler(void *token, int addr, int dummy0,
			    int dummy1, int dummy2) {
  free((unsigned char *)addr);
  /* Send a reply to the null function on the requestor. */
  AM_Safe(AM_Reply4(token, bulkNullHandlerIdx, 0, 0, 0, 0));
}

/* The wonderful An-Active-Message-Request-Handler-Must-Make-A-Reply
   workaround. */
static void bulkNullHandler(void *token, int dummy0, int dummy1,
			    int dummy2, int dummy3) {
  return;
}

void bulkSetupHandlers() {
  /* These had better map to the same handler indices on all processes, or
     I'm gonna have a conniption. */
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &bulkGather2HandlerIdx, 
			   bulkGather2Handler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &bulkAllocHandlerIdx, 
		   bulkAllocHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &bulkAllocReplyHandlerIdx, 
		   bulkAllocReplyHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &bulkFreeHandlerIdx, 
			   bulkFreeHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &bulkNullHandlerIdx, 
			   bulkNullHandler));
  AM_Safe(AM_SetHandlerAny(__sc_endpoint, &bulkScatter2HandlerIdx, 
			   bulkScatter2Handler));
}
