#ifndef __REGION
#define __REGION

/* This implementation assumes 32 bit pointers in the use of private
   region addresses as RegionId's and in calls to __i_read to read
   remote pointers (Split-C backend) */

#include "runtime-options.h"

#include <stdlib.h>
#include "linkage.h"
#include "region-struct.h"
#include "ti-memory.h"

typedef struct cl_ti_lang_Region *Region;

struct cl_ti_lang_SharedRegion;
struct cl_ti_lang_PrivateRegion;

#define RPAGESIZE 8192
#define RPAGELOG 13

#define PAGENB(x) ((uint32)(x) >> RPAGELOG)
#define PAGEADDR(n) ((void *)((n) << RPAGELOG))


#ifdef WIDE_POINTERS
typedef struct {
  int proc;
  Region addr;
} GRegion;

typedef struct {
  int proc;
  uint32 *addr;
} GP_uint32;

#else
typedef Region GRegion;
#endif


#define FREEPAGE ((RegionId)-1) /* Id of a free page */
#ifdef NMEMDEBUG
#define ASSERT_FREE(p) 0
#define ASSERT_INUSE(p, r) 0
#else
#define ASSERT_FREE(p) assert(__regionmap[PAGENB((p))] == FREEPAGE)
#define ASSERT_INUSE(p, r) assert(__regionmap[PAGENB((p))] == Region2RegionId((r)))
#endif

#ifndef BACKEND_UNIPROC
typedef long RegionId;
#define GCREGIONID 1
#else  /* BACKEND_UNIPROC */
/* regions and region ids are the same thing */
typedef Region RegionId; 
extern struct cl_ti_lang_Region zeroregion;
#define GCREGIONID &zeroregion
#endif /* BACKEND_UNIPROC */

EXTERN_C_BEGIN

void region_init(void);
struct cl_ti_lang_SharedRegion *newsregion(void);
struct cl_ti_lang_PrivateRegion *newpregion(void);
Region newregion(void);
void initregion(Region r);
void *ralloc(Region r, size_t size);
const void *ralloc_atomic(Region r, size_t size);
int deleteregion(Region r);

EXTERN_C_END

#define MAXMEMBITS 32
#define MAXPAGE (1 << (MAXMEMBITS - RPAGELOG))

extern RegionId __regionmap[];

static INLINE_MODIFIER RegionId PL2RegionId(void *p)
{
  unsigned long pagenb = PAGENB(p);

  return __regionmap[pagenb];
}

#ifdef MEMORY_DISTRIBUTED

extern Region __sregionidmap[];

static INLINE_MODIFIER Region PL2Region(void *p)
{
  RegionId rid = PL2RegionId(p);

  if (rid & 1) return __sregionidmap[rid >> 1];
  else return (Region)rid;
}

# define SHARED_REGIONID(rid) ((rid) & 1)

# define Region2RegionId(r) ((r)->id)

# define MyRegion(r) (r)

static INLINE_MODIFIER int GRegion2RegionId(GRegion r)
{
  uint32 id;
  GP_uint32 idAddr;

  FIELD_ADDR_GLOBAL(idAddr, r, id);
  DEREF_GLOBAL(id, idAddr);
  return id;
}

# define PG2RegionId(p)  split2RegionId((p).proc, (p).addr)

static INLINE_MODIFIER RegionId
split2RegionId(int proc, const void *addr)
{
  unsigned long pagenb = PAGENB(addr);

  if (proc == MYPROC || !addr) return __regionmap[pagenb];
  else return __i_read(___create_global_ptr(proc, __regionmap + pagenb));
}

# define PG2Region(p)  split2Region((p).proc, (p).addr)

static INLINE_MODIFIER Region
split2Region(int proc, const void *addr)
{
  RegionId rid = split2RegionId(proc, addr);

  if (rid & 1) return __sregionidmap[rid >> 1];
  else return (Region)rid;
}

static INLINE_MODIFIER Region MyGRegion(r)
GRegion r;
{
  if (isLocal(r)) return (Region)TO_LOCAL(r);
  /* Must be a shared region (otherwise we wouldn't be seeing it) */
  return __sregionidmap[GRegion2RegionId(r) >> 1];
}

#else /* !MEMORY_DISTRIBUTED */
/* uniproc, smp-narrow, smp-wide, and pthread backends */

# ifdef MEMORY_SHARED

extern Region __sregionidmap[][MAX_SMP_PROCS];

static INLINE_MODIFIER Region PL2Region(void *p)
{
  unsigned long pagenb = PAGENB(p);
  RegionId rid;

  rid = __regionmap[pagenb];
  if (rid & 1) return __sregionidmap[rid >> 1][MYPROC];
  else return (Region)rid;
}

#  define SHARED_REGIONID(rid) ((rid) & 1)

#  define Region2RegionId(r) ((r)->id)

static INLINE_MODIFIER Region MyRegion(Region r)
{
  RegionId rid;

  if (!r) return NULL;
  /* Returns our representative for region r */
  rid = r->id;

  if (rid & 1) return __sregionidmap[rid >> 1][MYPROC];
  else return r;
}

# else  /* !MEMORY_SHARED */

/* uniproc backend */
#  define PL2Region PL2RegionId
#  define SHARED_REGIONID(rid) ((rid)->id & 1)
#  define Region2RegionId(r) (r)
#  define MyRegion(r) (r)

# endif /* !MEMORY_SHARED */

# ifdef WIDE_POINTERS
#  define GRegion2RegionId(x) Region2RegionId((Region)TO_LOCAL(x))
#  define PG2RegionId(x) PL2RegionId((Region)TO_LOCAL(x))
#  define PG2Region(x) PL2Region((Region)TO_LOCAL(x))
#  define MyGRegion(x) MyRegion((Region)TO_LOCAL(x))
# else /* !WIDE_POINTERS */
#  define GRegion2RegionId Region2RegionId
#  define PG2RegionId PL2RegionId
#  define PG2Region PL2Region
#  define MyGRegion(x) MyRegion((GRegion)(x))
# endif/* !WIDE_POINTERS */

#endif /* !MEMORY_DISTRIBUTED */

#endif /* !__REGION */
