// Resolve operator overloading, rewrite as method invocations

#include "AST.h"
#include "decls.h"
#include "strings.h"

TreeNode *TreeNode::generalOverload(string op, bool assign,
				    TreeNode *object, TreeListNode *args)
{
  TypeNode *otype = object->type();

  if (!(otype->kind() & (ClassKind | InterfaceKind | ImmutableKind)))
    return this;

  const string *opString = intern(op);
  EnvironIter methods = otype->decl()->environ()
    ->lookupFirstProper(opString, Decl::Method);
  if (methods.isDone()) return this; // leave error to type checker

  SourcePosn p = position();
  Decl *method = resolveCall(methods, (otype->modifiers() & Local) != 0,
			     args, p);

  // build a method call node
  TreeNode *methodAccess =
    new ObjectFieldAccessNode(object, new NameNode(TreeNode::omitted, opString, method, p), p);

  TreeNode *methodNode;
  if (assign)
    methodNode = new MethodCallAssignNode(methodAccess, NULL, p);
  else
    methodNode = new MethodCallNode(methodAccess, NULL, p);
  methodNode->args(args);

  return methodNode;
}

TreeNode *TreeNode::overload(bool assign)
{
  TreeListNode *methodArgs =
    new TreeListNode((llist<TreeNode *> *)(arity() == 1 ? NULL : cons(child(1))),
		     position());
  return generalOverload(operatorName(), assign, child(0), methodArgs);
}

TreeNode *TreeNode::overloadArray()
{
  return generalOverload("[]", false, array(), index()->args());
}

TreeNode *TreeNode::overloadArrayAssign()
{
  llist<TreeNode *> *mlArgs = appendTreeList(opnd0()->index()->args(),
					     cons(opnd1()));
  TreeListNode *methodArgs = new TreeListNode(mlArgs, position());
  return generalOverload("[]=", false, opnd0()->array(), methodArgs);
}

TreeNode *TreeNode::_resolveOperators()
{
  return this;
}

TreeNode *ComplementNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *NotNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *UnaryArithNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *BinaryArithNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *PlusNode::_resolveOperators()
{
  return overload( false );
}

TreeNode *ShiftNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *RelationNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *EqualityNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *BitwiseNode::_resolveOperators()
{
  return overload(false);
}

TreeNode *BinaryArithAssignNode::_resolveOperators()
{
  return overload(true);
}

TreeNode *ShiftAssignNode::_resolveOperators()
{
  return overload(true);
}

TreeNode *BitwiseAssignNode::_resolveOperators()
{
  return overload(true);
}

// operatorName (this is used for more than just operator overloading, so
// it is defined on all operators)

const char *TreeNode::operatorName() const
{
  undefined("operatorName");
  return NULL;
}

const char *ArrayAccessNode::operatorName() const
{
  return "[]";
}

const char *ComplementNode::operatorName() const
{
  return "~";
}

const char *NotNode::operatorName() const
{
  return "!";
}

const char *PostIncrNode::operatorName() const
{
  return "++";
}

const char *PostDecrNode::operatorName() const
{
  return "--";
}

const char *PreIncrNode::operatorName() const
{
  return "++";
}

const char *PreDecrNode::operatorName() const
{
  return "--";
}

const char *UnaryPlusNode::operatorName() const
{
  return "+";
}

const char *UnaryMinusNode::operatorName() const
{
  return "-";
}

const char *MultNode::operatorName() const
{
  return "*";
}

const char *DivNode::operatorName() const
{
  return "/";
}

const char *RemNode::operatorName() const
{
  return "%";
}

const char *PlusNode::operatorName() const
{
  return "+";
}

const char *MinusNode::operatorName() const
{
  return "-";
}

const char *LeftShiftLogNode::operatorName() const
{
  return "<<";
}

const char *RightShiftLogNode::operatorName() const
{
  return ">>>";
}

const char *RightShiftArithNode::operatorName() const
{
  return ">>";
}

const char *LTNode::operatorName() const
{
  return "<";
}

const char *GTNode::operatorName() const
{
  return ">";
}

const char *LENode::operatorName() const
{
  return "<=";
}

const char *GENode::operatorName() const
{
  return ">=";
}

const char *EQNode::operatorName() const
{
  return "==";
}

const char *NENode::operatorName() const
{
  return "!=";
}

const char *BitAndNode::operatorName() const
{
  return "&";
}

const char *BitOrNode::operatorName() const
{
  return "|";
}

const char *BitXorNode::operatorName() const
{
  return "^";
}

const char *CandNode::operatorName() const
{
  return "&&";
}

const char *CorNode::operatorName() const
{
  return "||";
}

const char *IfExprNode::operatorName() const
{
  return "?:";
}

const char *AssignNode::operatorName() const
{
  return "=";
}

const char *PlusAssignNode::operatorName() const
{
  return "+=";
}

const char *MultAssignNode::operatorName() const
{
  return "*=";
}

const char *DivAssignNode::operatorName() const
{
  return "/=";
}

const char *RemAssignNode::operatorName() const
{
  return "%=";
}

const char *MinusAssignNode::operatorName() const
{
  return "-=";
}

const char *LeftShiftLogAssignNode::operatorName() const
{
  return "<<=";
}

const char *RightShiftLogAssignNode::operatorName() const
{
  return ">>>=";
}

const char *RightShiftArithAssignNode::operatorName() const
{
  return ">>=";
}

const char *BitAndAssignNode::operatorName() const
{
  return "&=";
}

const char *BitOrAssignNode::operatorName() const
{
  return "|=";
}

const char *BitXorAssignNode::operatorName() const
{
  return "^=";
}

const char *StringConcatNode::operatorName() const
{
  return "string concatenation";
}

const char *StringConcatAssignNode::operatorName() const
{
  return "assigned string concatenation";
}

const char *PointNode::operatorName() const
{
  return "point expressions";
}

const char *DomainNode::operatorName() const
{
  return "domain expressions";
}

const char *InstanceOfNode::operatorName() const
{
  return "instanceof";
}

const char *CastNode::operatorName() const
{
  return "cast";
}

// Operator names for statements (for error messages only)

const char *WhileNode::operatorName() const
{
    return "while";
}

const char *DoNode::operatorName() const
{
    return "do-while";
}

const char *ForNode::operatorName() const
{
  return "for";
}

const char *SwitchNode::operatorName() const
{
  return "switch";
}

const char *SwitchBranchNode::operatorName() const
{
  return "switch-branch";
}

const char *ReturnNode::operatorName() const
{
  return "return";
}

const char *IfStmtNode::operatorName() const
{
  return "if";
}

const char *ThrowNode::operatorName() const
{
  return "throw";
}

const char *SynchronizedNode::operatorName() const
{
  return "synchronized";
}

const char *CatchNode::operatorName() const
{
  return "try-catch";
}

const char *TryStmtNode::operatorName() const
{
  return "try";
}

const char *ForEachStmtNode::operatorName() const
{
  return "foreach";
}

const char *PartitionClauseNode::operatorName() const
{
  return "partition clauses";
}

const char *BroadcastNode::operatorName() const
{
  return "broadcast";
}
