#include "utils.h"
#include "AST.h"
#include "decls.h"
#include "errors.h"

void TreeNode::resolvePackage(Decl*, Environ *)
{
  undefined("resolvePackage");
}

void CompileUnitNode::resolvePackage(Decl *, Environ *)
{
  foriter (type, types()->allChildren(), ChildIter)
    (*type)->resolvePackage (thePackage, environ);

  // Must do this after the types, because we want to build this files environment
  // before we start loading the imported types
  foriter (import, imports()->allChildren(), ChildIter)
    (*import)->resolveImports(this, environ);
}

void ClassDeclNode::resolvePackage(Decl *package, Environ *fileEnv)
{
  Decl *other = fileEnv->lookupProper(simpName()->ident());

  if (other)
    {
      Error ("attempt to redefine " + other->errorName() + " as a class",
	     position());
      return;
    }

  Environ *packageEnv = fileEnv->parent();

  other = packageEnv->lookupProper(simpName()->ident());
  if (other && (!other->source() || other->source()->absent()))
    {
      // Assume this is the definition of 'other'
      other->source(this);
      other->modifiers(flags());
    }
  else
    {
      ClassDecl *cl = new ClassDecl(simpName()->ident(), package, flags(), this);

      if (other) // Redefinition in same package.
	Error ("class name " + *simpName()->ident() +
	       " conflicts with " + other->errorName() +
	       " in same package", position());
      else
	packageEnv->add(cl);

      other = cl;
    }
  other->environ(new Environ(fileEnv));
  fileEnv->add(other);

  simpName()->decl(other);
}

void InterfaceDeclNode::resolvePackage(Decl *package, Environ *fileEnv)
{
  Decl *other = fileEnv->lookupProper(simpName()->ident());

  if (other)
    {
      Error ("attempt to redefine " + other->errorName() + " as an interface",
	     position());
      return;
    }

  Environ *packageEnv = fileEnv->parent();

  other = packageEnv->lookupProper(simpName()->ident());
  if (other && (!other->source() || other->source()->absent()))
    {
      // Assume this is the definition of 'other'
      other->source(this);
      other->modifiers((Common::Modifiers)(Common::Interface | flags()));
    }
  else
    {
      ClassDecl *cl = new ClassDecl(simpName()->ident(), package,
				    (Common::Modifiers) (Common::Interface | flags()),
				    this);

      if (other) // Redefinition in same package.
	Error ("interface name " + *simpName()->ident() +
	       " conflicts with " + other->errorName() +
	       " in same package", position());
      else
	packageEnv->add(cl);

      other = cl;
    }
  other->environ(new Environ(fileEnv));
  fileEnv->add(other);

  simpName()->decl(other);
}

void TreeNode::resolveImports(CompileUnitNode *, Environ *)
{
  undefined("resolveImports");
}

void ImportNode::resolveImports(CompileUnitNode *, Environ *fileEnv)
{
  int errs0 = NumErrors();
  resolveAName(name(), *(PackageDecl::System->environ()), NULL, false, NULL,
	       Decl::Class | Decl::Interface);
  if (errs0 == NumErrors())
    {
      Decl *old = fileEnv->lookupProper(name()->ident());
      if (old != NULL && old != name()->decl())
	{
	  if (old != name()->decl())
	    Error ("attempt to import conflicting name: " + old->errorName(),
		   position());
	}
      else
	fileEnv->add(name()->decl());
    }
}

void ImportOnDemandNode::resolveImports(CompileUnitNode *file, Environ *)
{
  int errs0 = NumErrors();
  resolveAName(name(), *(PackageDecl::System->environ()), NULL, false, NULL,
	       Decl::Package);
  if (errs0 == NumErrors())
    file->importOnDemand(name()->decl());
}

void CompileUnitNode::importOnDemand(Decl *importedPackage)
{
  // ignore duplicate imports
  if (find(importedPackage, importedPackages)) return;
  importedPackages = cons(importedPackage, importedPackages);

  Environ* env = environ->parent()->parent();
  foriter (type, importedPackage->environ()->allProperDecls(), EnvironIter)
    if (type->category() != Decl::Package)
      env->add(&*type); // conflicts appear on use only
}

void CompileUnitNode::importOnDemand(const string *s1, const string *s2)
{
  TreeNode* name =
    new NameNode(new NameNode(TreeNode::omitted, s1, NULL), s2, NULL);

  int errs0 = NumErrors();
  resolveAName(name, *(PackageDecl::System->environ()), NULL, false, NULL,
	       Decl::Package);

  if (errs0 == NumErrors())
    importOnDemand(name->decl());
}

// Load all the explicitly visible types (more might appear when we
// start resolving names, but we don't need those to build the class-level
// environment
void TreeNode::resolveTypes(Decl *package, Environ *fileEnv)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->resolveTypes(package, fileEnv);
}

void TypeNameNode::resolveTypes(Decl *package, Environ *fileEnv)
{
  name(resolveAName(name(), *fileEnv, NULL, false, package, Decl::Class | Decl::Interface));
}

void CompileUnitNode::resolveTypes(Decl *, Environ *)
{
  foriter (type, types()->allChildren(), ChildIter)
    (*type)->resolveTypes(thePackage, environ);
}

void TreeNode::packageResolution() 
{
  undefined("packageResolution");
}

void CompileUnitNode::packageResolution() 
{
  if (package()->absent())
    thePackage = unnamedPackage;
  else
    thePackage = resolveAName(package(), *PackageDecl::System->environ(),
			      NULL, false, NULL, Decl::Package)->decl();

  // build environment for this file
  Environ *importOnDemandEnv = new Environ(PackageDecl::System->environ());
  Environ *packageEnv = new Environ(importOnDemandEnv);
  environ = new Environ(packageEnv); // the file level environment
  packageEnv->copy(thePackage->environ());

  importOnDemand(JavaString, LangString);
  importOnDemand(TitaniumString, LangString);
  resolvePackage(NULL, NULL);
  resolveTypes(NULL, NULL);
}
