#include "AST.h"
#include "UniqueId.h"
#include "code-util.h"
#include "compiler.h"
#include "decls.h"
#include "delimit.h"
#include "st-field.h"
#include "threading.h"


/*

The intermediate form (IF) we are constructing has a simplified form
compared to what the frontend creates and manipulates.

INPUT: a correct AST.  A correct AST has parent pointers set correctly
	and has types on all expressions, among other things.

OUTPUT: IF, a correct AST that conforms to the rules below.


All expressions must be "baby expressions" of the following forms:
(this should be expressed in terms of AST nodes rather than
pseudo code)

1.  var
2a. var.field
2b. var.i0.i1....ik.field	(where i0, ..., ik are immutable)
3a. op var			(unary op)
3b. var op var			(binary op)
3c. var.method(var, var, ...)	(method call)
3d. method(var, var, ...)	(static method call)
4.  var[var]			(array access)
5.  (type) var			(cast)
6.  ibroadcast var from var	(see note below)

In the above, var means a local variable, compiler generated
temporary, or constant.  In practice, it will often be a temporary
variable that is written once and read once, but not always.  In "baby
expressions" that contain multiple "vars," the "vars" may be evaluated
in any order.  Rewriting to enforce ordering constraints specified by
the source language must take place before/during the conversion to IF.

In Titanium, "broadcast foo from bar" has special evaluations rules.
In the IF, evaluation (on all procs) of the sending proc and
evaluation (on the sending proc) of the value to be sent must precede
the ibroadcast node.  We could have used the same AST node type
(broadcast), but by having two different node types there should be
less confusion.  The type of ibroadcast is (T, int) -> T, where T is
any type, so we can't rewrite it as a normal method call.

An assignment must be an AssignNode of the following form:

e = var		(where e is a baby expression and is assignable from var)

All control flow must be expressed using method calls or the following
node types:

IfStmtNode
WhileNode
DoNode
ForEachStmtNode (with ForEachPairNode)
SwitchNode (with CaseNode and SwitchBranchNode)
BreakNode, ContinueNode, LabeledStmtNode
ReturnNode
SynchronizedNode
TryStmtNode, TryNode, CatchNode, FinallyNode, ThrowNode
BlockNode

*/


TreeNode *CloneTree(TreeNode *t)
{
    TreeNode *newNode = t->clone();
    const int childCount = newNode->arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	newNode->child( sweep, CloneTree(newNode->child(sweep)) );

    return newNode;
}

BlockNode *AddSomething(BlockNode *b, TreeNode *n)
{
    TreeListNode *s = b->stmts();
    int c = s->arity();
    llist<TreeNode *> *t = NULL;
    for (int i=c-1; i>=0; i--)
	t = cons(s->child(i), t);
    
    b->stmts(new TreeListNode(cons(n, t)));
    return b;
}

BlockNode *AddAssignStmt(BlockNode *b, Decl *d, char *s)
{
    TreeNode *valNode = new LitNode (intLiteral(string("99")));
    TreeNode *varNode = new ObjectNode (new NameNode(TreeNode::omitted, intern(s), d));
    TreeNode *ass = new AssignNode(varNode, valNode);
    TreeNode *stmt = new ExpressionStmtNode(ass);
    return AddSomething(b, stmt);
}

BlockNode *CheckAddVariable(BlockNode *b, char *s)
{
    const string *str = intern(s);
    TreeNode *name = new NameNode (TreeNode::omitted, str, (Decl *)NULL);
    TreeNode *varDecl = new VarDeclNode (theIntType, name, TreeNode::omitted);
    Decl *d = new LocalVarDecl(varDecl->simpName()->ident(), varDecl->dtype(), varDecl);
    varDecl->simpName()->decl(d);
    b = AddAssignStmt(b, d, s);
    return AddSomething(b, varDecl);
}

static UniqueId tempGenerator( "___temp" );

TreeNode *MakeTemporary(TreeNode *expr, TreeNode *&declStmt, TreeNode *&assnStmt)
{
    const string *str = intern(tempGenerator.next().c_str());
    TreeNode *name = new NameNode (TreeNode::omitted, str, (Decl *)NULL);
    TreeNode *varDecl = new VarDeclNode (expr->type(), name, TreeNode::omitted);
    Decl *d = new LocalVarDecl(varDecl->simpName()->ident(), varDecl->dtype(), varDecl);
    varDecl->simpName()->decl(d);
    declStmt = varDecl;

    TreeNode *varNode = new ObjectNode (new NameNode(TreeNode::omitted, str, d));
    assnStmt = new ExpressionStmtNode(new AssignNode(varNode, expr));
    return new ObjectNode (new NameNode(TreeNode::omitted, str, d));
}

BlockNode *AddBlockStmts(BlockNode *b, TreeNode *insertPt,
			 llist<TreeNode *> *newStmts)
{
    TreeListNode *s = b->stmts();
    int c = s->arity();
    llist<TreeNode *> *t = NULL;
    if (c != 0) {
	for (int i=c-1; i>=0; i--) {
	    t = cons(s->child(i), t);
	    if (s->child(i) == insertPt) {
		for ( ; newStmts; newStmts = newStmts->tail() ) {
		    t = cons(newStmts->front(), t);
		}
	    }
	}
    } else {
	for ( ; newStmts; newStmts = newStmts->tail() )
	    t = cons(newStmts->front(), t);
    }
    b->stmts(new TreeListNode(t));
    return b;
}

SwitchBranchNode *AddSwitchBranchStmts(SwitchBranchNode *b, TreeNode *insertPt,
				       llist<TreeNode *> *newStmts)
{
    TreeListNode *s = b->stmts();
    int c = s->arity();
    llist<TreeNode *> *t = NULL;
    if (c != 0) {
	for (int i=c-1; i>=0; i--) {
	    t = cons(s->child(i), t);
	    if (s->child(i) == insertPt) {
		for ( ; newStmts; newStmts = newStmts->tail() ) {
		    t = cons(newStmts->front(), t);
		}
	    }
	}
    } else {
	for ( ; newStmts; newStmts = newStmts->tail() )
	    t = cons(newStmts->front(), t);
    }
    b->stmts(new TreeListNode(t));
    return b;
}

void AddStmts(TreeNode *insertPt, llist<TreeNode *> *newStmts)
{
    TreeNode *container = insertPt->parent()->parent();
    if (strcmp(container->oper_name(), "BlockNode") == 0) {
	AddBlockStmts((BlockNode *)container, insertPt, newStmts);
    } else if (strcmp(container->oper_name(), "SwitchBranchNode") == 0) {
	AddSwitchBranchStmts((SwitchBranchNode *)container, insertPt, newStmts);
    }
}

// newStmts is in reverse order...
BlockNode *AddBlockStmtsAtBegin(BlockNode *b, llist<TreeNode *> *newStmts)
{
    TreeNode *insert = (b->stmts()->arity() ? b->stmts()->child(0) : (TreeNode *)NULL);
    return AddBlockStmts(b, insert, newStmts);
}

TreeNode *TreeNode::introduceTemporary(TreeNode *func,
				       llist<TreeNode *> *&decls,
				       llist<TreeNode *> *&stmts)
{
    const int childCount = arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	child( sweep, child( sweep )->introduceTemporary(func, decls, stmts));

    return this;
}

TreeNode *ExpressionStmtNode::introduceTemporary(TreeNode *func,
						 llist<TreeNode *> *&decls,
						 llist<TreeNode *> *&stmts)
{
    stmts = NULL;
    child(0, child(0)->introduceTemporary(func, decls, stmts));
    AddStmts(this, stmts);
    stmts = NULL;
    return this;
}

TreeNode *ReturnNode::introduceTemporary(TreeNode *func,
					 llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
    stmts = NULL;
    child(0, child(0)->introduceTemporary(func, decls, stmts));
    AddStmts(this, stmts);
    stmts = NULL;
    return this;
}

TreeNode *ThrowNode::introduceTemporary(TreeNode *func,
					 llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
    stmts = NULL;
    child(0, child(0)->introduceTemporary(func, decls, stmts));
    AddStmts(this, stmts);
    stmts = NULL;
    return this;
}

TreeNode *SynchronizedNode::introduceTemporary(TreeNode *,
					       llist<TreeNode *> *&decls,
					       llist<TreeNode *> *&stmts)
{
    decls = NULL; stmts = NULL;
    if (stmt() != TreeNode::omitted) {
      if (!isBlockNode(stmt())) {
	stmt( stmt()->introduceTemporary(this, decls, stmts) );
      }
      else {
	stmt( stmt()->introduceTemporary(this, decls, stmts) );
	stmt( AddBlockStmtsAtBegin((BlockNode *)stmt(), stmts) );
	stmt( AddBlockStmtsAtBegin((BlockNode *)stmt(), decls) );
      }
      decls = NULL; stmts = NULL;
    }    
    return this;
}

TreeNode *MethodDeclNode::introduceTemporary(TreeNode *,
					     llist<TreeNode *> *&decls,
					     llist<TreeNode *> *&stmts)
{
  decls = NULL; stmts = NULL;
  if (body() != TreeNode::omitted) {
    if (!isBlockNode(body())) {
      body( body()->introduceTemporary(this, decls, stmts) );
    }
    else {
      body( body()->introduceTemporary(this, decls, stmts) );
      body( AddBlockStmtsAtBegin((BlockNode *)body(), stmts) );
      body( AddBlockStmtsAtBegin((BlockNode *)body(), decls) );
    }
    decls = NULL; stmts = NULL;
  }    
  return this;
}

TreeNode *ConstructorDeclNode::introduceTemporary(TreeNode *,
						  llist<TreeNode *> *&decls,
						  llist<TreeNode *> *&stmts)
{
    decls = NULL; stmts = NULL;
    if (body() != TreeNode::omitted) {
      if (!isBlockNode(body())) {
	body( body()->introduceTemporary(this, decls, stmts) );
      }
      else {
	body( body()->introduceTemporary(this, decls, stmts) );
	body( AddBlockStmtsAtBegin((BlockNode *)body(), stmts) );
	body( AddBlockStmtsAtBegin((BlockNode *)body(), decls) );
      }
      decls = NULL; stmts = NULL;
    }    
    return this;
}

void SortVarDeclsInBlock(BlockNode *b)
{
    TreeListNode *s = b->stmts();
    llist<TreeNode *> *decls = NULL, *newStmts = NULL;
    int c = s->arity();
    for (int i=c-1; i>=0; i--) {
	TreeNode *t = s->child(i);
	if (strcmp(t->oper_name(), "VarDeclNode") == 0) {
	    VarDeclNode *d = (VarDeclNode *)t;

	    if (!d->initExpr()->absent()) {

		TreeNode *init = d->initExpr();
		d->initExpr(TreeNode::omitted);

		TreeNode *varNode = new ObjectNode (d->simpName()->clone());
		TreeNode *assn = new ExpressionStmtNode(new AssignNode(varNode, init));
		newStmts = cons(assn, newStmts);
	    }
	    decls = cons(t, decls);
	} else {
	    newStmts = cons(t, newStmts);
	}
    }
    b->stmts(new TreeListNode(extend(decls, newStmts)));
}

void SortVarDeclsInSwitch(SwitchBranchNode *b)
{
    TreeListNode *s = b->stmts();
    llist<TreeNode *> *decls = NULL, *newStmts = NULL;
    int c = s->arity();
    for (int i=c-1; i>=0; i--) {
	TreeNode *t = s->child(i);
	if (strcmp(t->oper_name(), "VarDeclNode") == 0) {
	    VarDeclNode *d = (VarDeclNode *)t;

	    if (!d->initExpr()->absent()) {

		TreeNode *init = d->initExpr();
		d->initExpr(TreeNode::omitted);

		TreeNode *varNode = new ObjectNode (d->simpName()->clone());
		TreeNode *assn = new ExpressionStmtNode(new AssignNode(varNode, init));
		newStmts = cons(assn, newStmts);
	    }
	    decls = cons(t, decls);
	} else {
	    newStmts = cons(t, newStmts);
	}
    }
    b->stmts(new TreeListNode(extend(decls, newStmts)));
}

TreeNode *BlockNode::introduceTemporary(TreeNode *func,
					llist<TreeNode *> *&decls,
					llist<TreeNode *> *&newStmts)
{
    SortVarDeclsInBlock(this);
    TreeListNode *s = stmts();
    const int childCount = s->arity();

    for (int i=0, j=0; i<childCount; i++) {
	TreeNode *n = s->child(i);
	TreeNode *t = s->child(i)->introduceTemporary(func, decls, newStmts);
	while (stmts()->child(j) != n) j++;
	stmts()->child(j, t);
    }
    // Every statement in the block should have inserted its newStmts into the
    // block.
    // The following assertion causes g++ to die hard. Debugging one compiler
    // is enough for today; just leave this here.
    //assert((s->arity == 0) || (newStmts == NULL));
    return this;
}

TreeNode *IfStmtNode::introduceTemporary(TreeNode *func,
					 llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
    if (strcmp(thenPart()->oper_name(), "BlockNode") != 0) {
	thenPart(new BlockNode(cons(thenPart())));
    }

    if (strcmp(elsePart()->oper_name(), "BlockNode") != 0) {
	elsePart(new BlockNode(cons(elsePart())));
    }

    stmts = NULL;
    condition( condition()->introduceTemporary(func, decls, stmts) );
    AddStmts(this, stmts);
    stmts = NULL;
    
    thenPart( thenPart()->introduceTemporary(func, decls, stmts) );
    
    stmts = NULL;
    elsePart( elsePart()->introduceTemporary(func, decls, stmts) );
    stmts = NULL;
    return this;
}

// Change occurrences of "o" by "n" in this subtree
void ChangeJumpTargets(TreeNode *node, TreeNode *o, TreeNode *n)
{
    if (strcmp(node->oper_name(), "BreakNode") == 0) {
	BreakNode *t = (BreakNode *)node;
	if (t->destination() == o)
	    t->destination(n);
	return;
    }
    if (strcmp(node->oper_name(), "ContinueNode") == 0) {
	ContinueNode *t = (ContinueNode *)node;
	if (t->destination() == o)
	    t->destination(n);
	return;
    }
    const int childCount = node->arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	ChangeJumpTargets(node->child(sweep), o, n);
}

TreeNode *ForEachStmtNode::introduceTemporary(TreeNode *func,
					      llist<TreeNode *> *&decls,
					      llist<TreeNode *> *&stmts)
{
    if (strcmp(stmt()->oper_name(), "BlockNode") != 0) {
	stmt(new BlockNode(cons(stmt())));
    }

    // Anything other than vars() or stmt()?
    assert(arity() == 2);
    vars((TreeListNode *)vars()->introduceTemporary(func, decls, stmts));
    AddStmts(this, stmts);
    stmts = NULL;
    stmt(stmt()->introduceTemporary(func, decls, stmts));
    return this;
    /*
    const int childCount = arity();

    for (int sweep = 0; sweep < childCount; ++sweep)
	child( sweep, child( sweep )->introduceTemporary(func, decls, stmts)); */
}

TreeNode *SwitchNode::introduceTemporary(TreeNode *func,
					llist<TreeNode *> *&decls,
					llist<TreeNode *> *&newStmts)
{
    newStmts = NULL;
    expr( expr()->introduceTemporary(func, decls, newStmts) );
    AddStmts(this, newStmts);
    newStmts = NULL;

    TreeListNode *swB = switchBlocks();
    const int childCount = swB->arity();
    for (int sweep = 0; sweep < childCount; ++sweep) {
	SwitchBranchNode *b = (SwitchBranchNode *)swB->child(sweep);

	SortVarDeclsInSwitch(b);
	TreeListNode *s = b->stmts();
	int num = s->arity();

	for (int i=0, j=0; i<num; i++) {
	    TreeNode *n = s->child(i);
	    TreeNode *t = s->child(i)->introduceTemporary(func, decls, newStmts);
	    while (b->stmts()->child(j) != n) j++;
	    b->stmts()->child(j, t);
	}
	
    }
    return this;
}

TreeNode *TryStmtNode::introduceTemporary(TreeNode *func,
					  llist<TreeNode *> *&decls,
					  llist<TreeNode *> *&stmts)
{
    const int childCount = arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	child( sweep, child( sweep )->introduceTemporary(func, decls, stmts));

    return this;
}

TreeNode *WhileNode::introduceTemporary(TreeNode *func,
					llist<TreeNode *> *&decls,
					llist<TreeNode *> *&stmts)
{
    if (strcmp(stmt()->oper_name(), "BlockNode") != 0) {
	stmt(new BlockNode(cons(stmt())));
    }

    stmts = NULL;
    test( test()->introduceTemporary(func, decls, stmts) );

    llist<TreeNode *> *copy = NULL, *s = stmts;
    for ( ; s; s = s->tail() )
	copy = cons(CloneTree(s->front()), copy);

    AddStmts(this, stmts);

    stmts = NULL;
    stmt( stmt()->introduceTemporary(func, decls, stmts) );

    // Make a block containing the stmt(), and the copy stmts
    stmt( new BlockNode(cons(stmt(), copy)));
    return this;
}

TreeNode *DoNode::introduceTemporary(TreeNode *func,
				     llist<TreeNode *> *&decls,
				     llist<TreeNode *> *&stmts)
{
    if (strcmp(stmt()->oper_name(), "BlockNode") != 0) {
	stmt(new BlockNode(cons(stmt())));
    }
    
    stmt( stmt()->introduceTemporary(func, decls, stmts) );

    stmts = NULL;
    test( test()->introduceTemporary(func, decls, stmts) );
    stmt( new BlockNode(cons(stmt(), reverse(stmts))));
    stmts = NULL;

    return this;
}

TreeNode *ForNode::introduceTemporary(TreeNode *func,
				      llist<TreeNode *> *&decls,
				      llist<TreeNode *> *&stmts)
{
    // Attach the updStmts to the end of a new created block
    TreeListNode *updStmts = update();
    llist<TreeNode *> *updList = NULL;
    for (int c=updStmts->arity()-1; c>=0; c--)
	updList = cons(updStmts->child(c), updList);
    updList = cons(stmt(), updList);
    TreeNode *loopBody = new BlockNode(updList);
    TreeNode *whileLoop;
    if (test()->absent())
	whileLoop = new WhileNode(new LitNode(Literal((bool)1)), loopBody);
    else
	whileLoop = new WhileNode(test(), loopBody);
    whileLoop->prepareBreak();
    ChangeJumpTargets(whileLoop, this, whileLoop);

    TreeListNode *initStmts = init();
    llist<TreeNode *> *t = cons(whileLoop);

    // Attach the initStmts to a new created enclosing block
    // with the whileLoop as the last statement of the block
    for (int c=initStmts->arity()-1; c>=0; c--)
	t = cons(initStmts->child(c), t);
    
    BlockNode *b = new BlockNode(t);
    return b->introduceTemporary(func, decls, stmts);
}

static TreeNode *ProcessExpr(TreeNode *expr,
			     TreeNode *func,
			     llist<TreeNode *> *&decls,
			     llist<TreeNode *> *&stmts)
{
    const int childCount = expr->arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	expr->child(sweep, expr->child(sweep)->introduceTemporary(func, decls, stmts));

    TreeNode *d, *s, *newExpr;
    newExpr = MakeTemporary(expr, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);
    return newExpr;
}


TreeNode *BinaryArithNode::introduceTemporary(TreeNode *func,
					      llist<TreeNode *> *&decls,
					      llist<TreeNode *> *&stmts)
{
    return ProcessExpr(this, func, decls, stmts);
}

TreeNode *AssignNode::introduceTemporary(TreeNode *func,
					 llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
    child(0, child(0)->introduceTemporary(func, decls, stmts));
    child(1, child(1)->introduceTemporary(func, decls, stmts));

    if (strcmp(parent()->oper_name(), "ExpressionStmtNode") == 0)
	return this;
    else {
	TreeNode *t = CloneTree(child(1));
	push( stmts, static_cast<TreeNode *>(new ExpressionStmtNode(this)) );
	return t;
    }
}

TreeNode *BinaryArithAssignNode::introduceTemporary(TreeNode *func,
						    llist<TreeNode *> *&decls,
						    llist<TreeNode *> *&stmts)
{
    TreeNode *c1 = child(0)->introduceTemporary(func, decls, stmts);
    TreeNode *c2 = child(1)->introduceTemporary(func, decls, stmts);
    TreeNode *t;
    if (strcmp(oper_name(), "MultAssignNode") == 0) {
	t = new MultNode(c1, c2);
    } else if (strcmp(oper_name(), "DivAssignNode") == 0) {
	t = new DivNode(c1, c2);
    } else if (strcmp(oper_name(), "RemAssignNode") == 0) {
	t = new RemNode(c1, c2);
    } else if (strcmp(oper_name(), "PlusAssignNode") == 0) {
	t = new PlusNode(c1, c2);
    } else if (strcmp(oper_name(), "MinusAssignNode") == 0) {
	t = new MinusNode(c1, c2);	
    } else if (strcmp(oper_name(), "LeftShiftLogAssignNode") == 0) {
	t = new LeftShiftLogNode(c1, c2);
    } else if (strcmp(oper_name(), "RightShiftLogAssignNode") == 0) {
	t = new RightShiftLogNode(c1, c2);
    } else if (strcmp(oper_name(), "RightShiftArithAssignNode") == 0) {
	t = new RightShiftArithNode(c1, c2);
    } else assert(0);
    // Temporary variable for rhs compute
    TreeNode *d, *s, *newExpr;
    newExpr = MakeTemporary(t, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);

    TreeNode *assign = new AssignNode(c1, newExpr);
    
    if (strcmp(parent()->oper_name(), "ExpressionStmtNode") == 0)
	return assign;
    else {
	TreeNode *t = CloneTree(newExpr);
	push(stmts, static_cast<TreeNode *>(new ExpressionStmtNode(assign)));
	return t;
    }
}

TreeNode *ShiftAssignNode::introduceTemporary(TreeNode *func,
					      llist<TreeNode *> *&decls,
					      llist<TreeNode *> *&stmts)
{
    TreeNode *c1 = child(0)->introduceTemporary(func, decls, stmts);
    TreeNode *c2 = child(1)->introduceTemporary(func, decls, stmts);
    TreeNode *t;

    if (strcmp(oper_name(), "LeftShiftLogAssignNode") == 0) {
	t = new LeftShiftLogNode(c1, c2);
    } else if (strcmp(oper_name(), "RightShiftLogAssignNode") == 0) {
	t = new RightShiftLogNode(c1, c2);
    } else if (strcmp(oper_name(), "RightShiftArithAssignNode") == 0) {
	t = new RightShiftArithNode(c1, c2);
    } else assert(0);
    // Temporary variable for rhs compute
    TreeNode *d, *s, *newExpr;
    newExpr = MakeTemporary(t, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);

    TreeNode *assign = new AssignNode(c1, newExpr);
    
    if (strcmp(parent()->oper_name(), "ExpressionStmtNode") == 0)
	return assign;
    else {
	TreeNode *t = CloneTree(newExpr);
	push(stmts, static_cast<TreeNode *>(new ExpressionStmtNode(assign)));
	return t;
    }
}

TreeNode *BitwiseAssignNode::introduceTemporary(TreeNode *func,
						llist<TreeNode *> *&decls,
						llist<TreeNode *> *&stmts)
{
    TreeNode *c1 = child(0)->introduceTemporary(func, decls, stmts);
    TreeNode *c2 = child(1)->introduceTemporary(func, decls, stmts);
    TreeNode *t;

    if (strcmp(oper_name(), "BitAndAssignNode") == 0) {
	t = new BitAndNode(c1, c2);
    } else if (strcmp(oper_name(), "BitXorAssignNode") == 0) {
	t = new BitXorNode(c1, c2);
    } else if (strcmp(oper_name(), "BitOrAssignNode") == 0) {
	t = new BitOrNode(c1, c2);
    } else assert(0);
    // Temporary variable for rhs compute
    TreeNode *d, *s, *newExpr;
    newExpr = MakeTemporary(t, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);

    TreeNode *assign = new AssignNode(c1, newExpr);
    
    if (strcmp(parent()->oper_name(), "ExpressionStmtNode") == 0)
	return assign;
    else {
	TreeNode *t = CloneTree(newExpr);
	push(stmts, static_cast<TreeNode *>(new ExpressionStmtNode(assign)));
	return t;
    }
}

TreeNode *FieldDeclNode::introduceTemporary(TreeNode *func,
					    llist<TreeNode *> *&decls,
					    llist<TreeNode *> *&stmts)
{
    return this;
}

TreeNode *StaticInitNode::introduceTemporary(TreeNode *func,
					     llist<TreeNode *> *&decls,
					     llist<TreeNode *> *&stmts)
{
    return this;
}

TreeNode *IfExprNode::introduceTemporary(TreeNode *func,
					 llist<TreeNode *> *&decls,
					 llist<TreeNode *> *&stmts)
{
//  TreeNode *cond = condition()->introduceTemporary(func, decls, stmts);
    return this;
}

TreeNode *MethodCallNode::introduceTemporary(TreeNode *func,
					     llist<TreeNode *> *&decls,
					     llist<TreeNode *> *&stmts)
{
    const int childCount = arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	child( sweep, child( sweep )->introduceTemporary(func, decls, stmts));

    Decl *dcl = method()->simpName()->decl();
    if (dcl->type()->returnType()->kind() != Common::VoidKind) {
	TreeNode *d, *s, *newExpr;
	newExpr = MakeTemporary(this, d, s);
	decls = cons(d, decls);
	stmts = cons(s, stmts);
	return newExpr;
    } else
	return this;    
}

TreeNode *IncrDecrNode::introduceTemporary(TreeNode *func,
					   llist<TreeNode *> *&decls,
					   llist<TreeNode *> *&stmts)
{
    TreeNode *t = child(0)->introduceTemporary(func, decls, stmts);
    TreeNode *d, *s, *old;

    old = MakeTemporary(t, d, s);
    decls = cons(d, decls);
    stmts = cons(s, stmts);

    TreeNode *oneNode = new LitNode (intLiteral(string("1")));
    TreeNode *assign;
    // Do the increment/decrement operation
    if ( (strcmp(oper_name(), "PostIncrNode") == 0)
	 || (strcmp(oper_name(), "PreIncrNode") == 0))
	assign = new AssignNode(t, new PlusNode(old, oneNode));
    else
	assign = new AssignNode(t, new MinusNode(old, oneNode));

    // If the order does not matter...
    if (strcmp(parent()->oper_name(), "ExpressionStmtNode") == 0)
	return assign;

    // The order matters.  So store the assignment operation.
    push(stmts, static_cast<TreeNode *>(new ExpressionStmtNode(assign)));

    // Return the old or the new value
    if ( (strcmp(oper_name(), "PostIncrNode") == 0)
	 || (strcmp(oper_name(), "PostDecrNode") == 0))
	return CloneTree(old);
    else
	return CloneTree(t);
}

static TreeNode *nodeMYPROC()
{
  TreeNode *m = new CodeLiteralExprNode("MYPROC");
  m->type(theIntType);
  return m;
}

/* Return an ExprNode that compares p to MYPROC. */
static TreeNode *isMYPROC(TreeNode *p)
{
  TreeNode *r = new EQNode(p, nodeMYPROC());
  r->type(theBoolType);
  return r;
}

TreeNode *BroadcastNode::introduceTemporary(TreeNode *func,
					    llist<TreeNode *> *&decls,
					    llist<TreeNode *> *&stmts)
{
  const SourcePosn pos = position();
  TreeNode *d1, *s1, *d2, *s2;
  TreeNode *p = MakeTemporary(proc()->introduceTemporary(func, decls, stmts),
			      d1, s1);
  llist<TreeNode *> *substmts = NULL;
  llist<TreeNode *> *subdecls = NULL;
  TreeNode *e =
    MakeTemporary(expr()->introduceTemporary(func, subdecls, substmts),
		  d2, s2);

  s2 = new BlockNode(cons(s2), pos);
  AddBlockStmtsAtBegin((BlockNode *) s2, substmts);
  AddBlockStmtsAtBegin((BlockNode *) s2, subdecls);

  TreeNode *ifMyBcastComputeExpr =
    new IfStmtNode(isMYPROC(p), s2, TreeNode::omitted, pos);

  return new BlockNode(cons(d1,
			    cons(d2,
				 cons(s1,
				      cons(ifMyBcastComputeExpr,
					   cons((TreeNode *)
						new IBroadcastNode(e,
								   p,
								   pos)))))),
		       pos);
}

TreeNode *AllocateNode::introduceTemporary(TreeNode *func,
					   llist<TreeNode *> *&decls,
					   llist<TreeNode *> *&stmts)
{
    return ProcessExpr(this, func, decls, stmts);
}

TreeNode *AllocateArrayNode::introduceTemporary(TreeNode *func,
					   llist<TreeNode *> *&decls,
					   llist<TreeNode *> *&stmts)
{
    return ProcessExpr(this, func, decls, stmts);
}

TreeNode *EqualityNode::introduceTemporary(TreeNode *func,
					   llist<TreeNode *> *&decls,
					   llist<TreeNode *> *&stmts)
{
    return ProcessExpr(this, func, decls, stmts);
}

TreeNode *LogCondNode::introduceTemporary(TreeNode *func,
					  llist<TreeNode *> *&decls,
					  llist<TreeNode *> *&stmts)
{
    return ProcessExpr(this, func, decls, stmts);
}

#if 0
TreeNode *TreeListNode::introduceTemporary(TreeNode *func,
					   llist<TreeNode *> *&decls,
					   llist<TreeNode *> *&stmts)
{
    const int childCount = arity();
    for (int sweep = 0; sweep < childCount; ++sweep)
	child(sweep, child(sweep)->introduceTemporary(func, decls, stmts));
    return this;
}

TreeNode *InstanceOfNode::introduceTemporary(TreeNode *,
					     llist<TreeNode *> *&decls,
					     llist<TreeNode *> *&stmts)
{
    return this;
}

TreeNode *ArrayAccessNode::introduceTemporary(TreeNode *func,
					      llist<TreeNode *> *&decls,
					      llist<TreeNode *> *&stmts)
{
    return this;    
}


TreeNode *ObjectFieldAccessNode::introduceTemporary(TreeNode *,
						    llist<TreeNode *> *&decls,
						    llist<TreeNode *> *&stmts)
{
    return this;
}

#endif

extern TreeNode *transform_init(TreeNode *t);

BlockNode *EncapsulateInits(vector<TreeNode *> inits)
{
    vector<TreeNode *>::iterator b = inits.begin();
    vector<TreeNode *>::iterator e = inits.end();
    llist<TreeNode *> *t = NULL;

    while (b < e) {
	t = cons(transform_init(*b++), t);
    }
    t = dreverse(t);
    BlockNode *block = new BlockNode(t);
    llist<TreeNode *> *d = NULL, *s = NULL;
    block->introduceTemporary(NULL, d, s);
    AddBlockStmtsAtBegin(block, s);
    AddBlockStmtsAtBegin(block, d);
    return block;
}

TreeNode *transform_init(TreeNode *t)
{
    if (strcmp(t->oper_name(), "StaticInitNode") == 0)
	return t;

    if (strcmp(t->oper_name(), "FieldDeclNode") == 0) {
	if (t->initExpr()->absent())
	    return t;
	TypeNode *cl = t->decl()->container()->asType();
	TreeNode *sname = new NameNode(TreeNode::omitted,
				       t->decl()->name(),
				       t->decl());
	TreeNode *fld = new ObjectFieldAccessNode(new ThisNode(cl, 1), sname);
	return new ExpressionStmtNode(new AssignNode (fld, t->initExpr()));
    }
    assert(0);
}



